"""
Task

This module contains actions related to the running task instance.
"""

from typing import Dict, List, Any

from ppa import task, ui


def manifest() -> Dict[str, Any]:
    """Get the task manifest.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"
    A dictionary containing the task manifest.

??? example
    ```YAML linenums="1"
    - ppa.task.manifest:
      save: manifest
    ```
"""
    return task.manifest()


def payload() -> Dict[str, Any]:
    """Get the task payload.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"
    A dictionary containing the task payload.

??? example
    ```YAML linenums="1"
    - ppa.task.payload:
      save: payload
    ```
"""
    return task.payload()


def get_payload_values(keys: List[str]) -> Dict[str, Any]:
    """Get specific keys & values from the task payload.

Only use this action if the task payload is a dictionary.

!!! question "Minimum Plugin Version: 3.1.0"

!!! tip "Missing Keys"

    This action will fail if any requested keys are missing.

??? info "Input"

    - __keys:__ a list of keys to get from the payload

??? success "Output"

    A dictionary containing the requested keys & values from the payload.

??? example
    ```YAML linenums="1"
    - ppa.task.get_payload_values:
        keys:
          - region_name
          - vpc_id
      save: payload
    ```
"""

    task_payload = payload()

    if not isinstance(task_payload, dict):
        raise TypeError("The task payload is not a dictionary.")

    if missing_keys := [key for key in keys if key not in task_payload]:
        message = "The following keys are missing from the task payload: '{}'.".format(", ".join(missing_keys))
        ui.output_error(message)
        raise KeyError(message)

    return {key: task_payload[key] for key in keys}


def user_password() -> str:
    """With the task operator's consent, get the password they used to authenticate to PPA.

This can be used to authenticate to other systems using the task operator's identity.

!!! question "Minimum Plugin Version: 1.0.0"

??? danger "Remember"

    This operation is not possible if SAML is used to authenticate users to PPA.

??? info "Input"

    This action takes no inputs.

??? success "Output"
    The operator's password if the consent was given, or `null` if it was not given.

??? example
    ```YAML linenums="1"
    - ppa.task.user_password:
      save: operator_password
    ```
"""
    return task.user_password()


def started_by_user() -> bool:
    """Determine if a task was started by a PPA user.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by a user, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_user:
      save: interactive
    ```
"""
    return task.trigger().lower() == "ui"


def started_by_schedule() -> bool:
    """Determine if a task was started by a PPA schedule.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by a schedule, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_schedule:
      save: scheduled
    ```
"""
    return task.trigger().lower() == "job"


def started_by_api() -> bool:
    """Determine if a task was started by an API call.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by an API call, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_api:
      save: api_triggered
    ```
"""
    return task.trigger().lower() == "api"
