"""
User Interface

This module contains actions for interacting with PPA Task Operators.

!!! question "PPA Task Interface"

    See [Task Interface](../task_interface) for more information on inputs & outputs in PPA.
"""

import functools
import uuid

from typing import Optional, List, Dict, Union, TypedDict, Any

import validators

from ppa import ui

from .records import UploadedFile


class OrderedDictChoice(TypedDict):
    value: Union[str, int]
    order: int


ChoiceList = List[Union[str, int]]
ChoiceSimpleDict = Dict[str, Union[str, int]]
ChoiceOrderedDict = Dict[str, OrderedDictChoice]
TableHeader = List[str]
TableRow = List[Union[str, int, bool, None]]
TableRows = List[TableRow]
TableTab = Dict[str, TableRows]


def input_text(
    text: str, default: Optional[str] = None, pattern: str = ".*", element_id: Optional[str] = None, required: bool = False,
) -> str:
    """Present a text input field.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the input field

    - __default:__ a default value for the input field (defaults to empty)

    - __pattern:__ a regular expression to validate the supplied value

    - __required:__ set to `true` to require a value before submission (defaults to `false`)

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    The string submitted by the task operator.

??? example
    Prompting for a username & saving it as a new variable called _username_.

    - The value is required, & must match the format _firstname.surname_
    
    - A folded string (the _>_ character) escapes any special pattern characters  
    
    ```YAML linenums="1"
    - ppa.ui.input_text:
        text: Username
        required: true
        pattern: >
          [A-Za-z]+\.[A-Za-z]+
      save: username
    ```
"""  # noqa
    return ui.input_text(text, default=default, pattern=pattern, element_id=element_id, required=required)


def input_multiline_text(
    text: str,
    default: Optional[str] = None,
    pattern: str = ".*",
    element_id: Optional[str] = None,
    required: bool = False,
    lines: int = 15,
) -> str:
    """Present a multiline text input field.

!!! question "Minimum Plugin Version: 3.4.0"

??? info "Input"

    - __text:__ the title of the input field

    - __default:__ a default value for the input field (defaults to empty)

    - __pattern:__ a regular expression to validate the supplied value

    - __required:__ set to `true` to require a value before submission (defaults to `false`)

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)
    
    - __lines:__ Denotes how many lines the input element should have (defaults to 15)

??? success "Output"
    The string submitted by the Task Operator. Any line breaks will be denoted by a `\n`.

??? example
    Prompting for a multiline comment & saving it as a new variable called _comment_.

    ```YAML linenums="1"
    - ppa.ui.input_multiline_text:
        text: Service Now Comment
        required: true
      save: comment
    ```
"""  # noqa
    return ui.input_text_area(
        text, default=default, pattern=pattern, element_id=element_id, required=required, rows=lines, max_rows=100
    )


def input_email(text: str, element_id: Optional[str] = None,) -> str:
    """Present a text input field used to supply an email address.

The address format will be validated, & the task operator will be re-prompted if validation fails.

!!! question "Minimum Plugin Version: 3.2.0"

??? info "Input"

    - __text:__ the title of the input field

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"

    The email address submitted by the task operator.

??? example

    ```YAML linenums="1"
    - ppa.ui.input_email:
        text: Please supply your email address
      save: email_address
    ```
"""
    while True:
        email_address = ui.input_text(text, element_id=element_id)
        if not validators.email(email_address):
            ui.output_error(f"Supplied email address '{email_address}' is not valid, please try again.")
            continue
        break
    return email_address


def input_choice(
    text: str, options: Union[ChoiceList, ChoiceSimpleDict, ChoiceOrderedDict], element_id: Optional[str] = None
) -> str:
    """Present a dropdown & wait for a choice.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the dropdown list

    - __options:__ one or more values to populate the dropdown list (see the examples for more information)

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    The string value chosen by the task operator.

??? example "Example 1: Simple"

    Prompting for a team & saving the choice as a new variable called _team_.

    ```YAML linenums="1"
    - ppa.ui.input_choice:
        text: Select a Team
        options:
          - Sales
          - Marketing
          - Engineering
      save: team
    ```

??? example "Example 2: Custom Ordering"

    Prompting for a quantity & saving the choice as a new variable called _quantity_

    - To avoid list values being automatically sorted, this example uses a dictionary as input
    - Each key is a dropdown item, & its value is its order in the list

    ```YAML linenums="1"
      - ppa.ui.input_choice:
          text: Select a Quantity
          options:
            One: 1
            Two: 2
            Three: 3
            Four: 4
            Five: 5
        save: quantity
    ```
"""

    @functools.singledispatch
    def _execute(_):
        raise Exception("The options passed to input_choice must be a list or dictionary.")

    @_execute.register(list)
    def _(options) -> str:
        return ui.input_choice(text, options=options, element_id=element_id)

    @_execute.register(dict)
    def _(options) -> str:
        # Is it a nested dict containing value and order keys? If any of the values is a
        # dict, it should resemble a ChoiceOrderedDict, otherwise an Exception is raised.
        if any([isinstance(value, dict) for _, value in options.items()]):
            if not all([sorted(list(inner.keys())) == ["order", "value"] for _, inner in options.items()]):
                raise Exception("Nested choice dictionaries must only contain 'order' and 'value' keys.")
            return options[
                ui.input_choice(
                    text, options=sorted(list(options.keys()), key=lambda key: options[key]["order"]), element_id=element_id,
                )
            ]["value"]

        # Otherwise it is a ChoiceSimpleDict and we automatically sort the choices by value.
        return options[
            ui.input_choice(text, options=sorted(list(options.keys()), key=lambda key: options[key]), element_id=element_id)
        ]

    return _execute(options)


def input_accept(text: str, element_id: Optional[str] = None) -> None:
    """Present a labelled button with text, & wait until it is clicked.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the button label
    - __element_id:__ the ID of the element in the task interface (see [here](../task_interface) for more)

??? success "Output"
    Nothing is returned from this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_accept:
        text: Start
    ```
"""
    ui.input_accept(text, element_id=element_id)


def input_password(text: str, meter: Optional[bool] = False, element_id: Optional[str] = None) -> str:
    """Present a password field.

The password will be obfuscated in the interface & not stored in PPA.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the input field

    - __meter:__ set to `true` to show a password strength meter (defaults to `false`)

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    The password supplied by the task operator.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_password:
        text: Enter a Password
        meter: true
      save: new_password
    ```
"""
    return ui.input_password(text, meter=meter, element_id=element_id)


def input_password_confirm() -> str:
    """Present 'New Password' & 'Confirm Password' input fields.

This action will loop until both supplied passwords match.

The passwords are obfuscated in the interface & not stored in PPA.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    This action takes no inputs.

??? success "Output"
    The password supplied by the task operator.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_password_confirm:
      save: new_password
    ```
"""
    new_password_id, confirm_password_id, error_id = list(map(lambda _: str(uuid.uuid4()), range(0, 3)))

    while True:
        new_password = ui.input_password("New Password", meter=False, element_id=new_password_id)
        confirm_password = ui.input_password("Confirm Password", meter=False, element_id=confirm_password_id)
        if new_password == confirm_password:
            return new_password
        else:
            ui.output_error("Passwords must match.", element_id=error_id)


def input_confirm(
    text: str, confirm: Optional[str] = None, cancel: Optional[str] = None, element_id: Optional[str] = None
) -> bool:
    """Present two labelled buttons to offer a choice, & wait until one is clicked.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the text to display above the 2 buttons

    - __confirm:__ the label of the button that returns `true` (defaults to _Confirm_)

    - __cancel:__ the label of the button that returns `false` (defaults to _Cancel_)

    - __element_id:__ an optional ID for the task interface element (see [here](../task_interface) for more)

??? success "Output"
    A boolean (`true` or `false`).

??? example
    ```YAML linenums="1"
    - ppa.ui.input_confirm:
        text: Would you like to continue?
        confirm: Yes
        cancel: No
      save: confirm
    ```
"""
    return ui.input_confirm(text, confirm, cancel, element_id=element_id)


def input_table(
    text: str,
    header: TableHeader,
    rows: TableRows,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    element_id: Optional[str] = None,
) -> List[List[Any]]:
    """Present a table & wait for a selection.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the table title

    - __header:__ a list of column names

    - __rows:__ the table contents as a list of lists, where each inner list is a row of values

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"

    A list of the selected table row indices (zero-indexed).

??? example

    Presenting a table of cities & countries & saving the selection as `selected_rows`.

    ```YAML linenums="1" hl_lines="13"
    - ppa.ui.input_table:
        title: Select Cities
        header:
          - City
          - Country
        rows:
          - - Paris
            - France
          - - London
            - England
          - - Cardiff
            - Wales
      save: selected_rows
    ```

    If the _first_ & _second_ rows were chosen `selected_rows` would be:

    ```YAML linenums="1"
      - 0
      - 1
    ```
"""
    while True:
        indices = ui.input_table(text, header=header, data=rows, element_id=element_id)
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=f"{element_id}_error")
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=f"{element_id}_error")
            continue
        break
    return indices


def input_file(text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None) -> UploadedFile:
    """Ask the Task Operator to upload a file to the PPA task container.

The file contents can be used & inspected by the task.

The file will be made available to be downloaded by the task operator.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __text:__ text to display with the upload component
    - __accept:__ optional file extension or MIME type to accept
    - __element_id:__ the ID of the element in the task interface (see [here](../task_interface) for more)

??? success "Output"
    A single [UploadedFile](../inputs_outputs/#uploadedfile) dictionary.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_file:
        text: Please upload a users CSV file
        accept: .csv
      save: users_csv
    ```
"""
    return UploadedFile(**ui.input_file(text, accept=accept, element_id=element_id)._asdict())


def input_sensitive_file(text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None) -> UploadedFile:
    """Ask the Task Operator to upload a file to the PPA task container.

The file contents can be used & inspected by the task.

The file will not be made available to be downloaded by the task operator.

!!! question "Minimum PPA Appliance Version: 2.6.3"

!!! question "Minimum Plugin Version: 3.4.0"

??? info "Input"

    - __text:__ text to display with the upload component
    - __accept:__ optional file extension or MIME type to accept
    - __element_id:__ the ID of the element in the task interface (see [here](../task_interface) for more)

??? success "Output"
    A single [UploadedFile](../inputs_outputs/#uploadedfile) dictionary.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_sensitive_file:
        text: Please upload a certificate private key
        accept: .pem
      save: private_key
    ```
"""
    return UploadedFile(**ui.input_sensitive_file(text, accept=accept, element_id=element_id)._asdict())


def output_info(text: str, element_id: Optional[str] = None) -> None:
    """Present some basic text.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the text to display

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.output_info:
        text: Welcome to PPA
    ```
"""
    ui.output_info(text, element_id=element_id)


def output_error(text: str, element_id: Optional[str] = None) -> None:
    """Present some error text in red.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the text to display

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.output_error:
        text: Something went wrong
    ```
"""
    ui.output_error(text, element_id=element_id)


def output_markdown(doc: str, element_id: Optional[str] = None) -> None:
    """Present a markdown document to display.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __doc:__ the markdown document to display

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.output_markdown:
        doc: >
          # Markdown Header

          Document text
    ```
"""
    ui.output_markdown(doc, element_id=element_id)


def output_progress(text: str, percent: Optional[int] = None, element_id: Optional[str] = None) -> None:
    """Present a progress bar with an indeterminate or % value.

The value _null_ is deemed indeterminate & will display an animation.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the progress bar

    - __percent:__ an integer progress value or `null`

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? tip "Progress Bar Element IDs"
    It is essential to supply an _element_id_ to any progress bar whose value will be set more than once.

??? success "Output"
    Nothing is outputted by this action.

??? example
    - Showing an indeterminate progress bar

    - waiting for a button click

    - Updating the progress bar to 100%

    ```YAML linenums="1"
    - ppa.ui.output_progress:
        text: Waiting to continue...
        percent: null
        element_id: action_progress
    - ppa.ui.input_accept:
        text: Continue
    - ppa.ui.output_progress:
        text: Continuing
        percent: 100
        element_id: action_progress
    ```
"""
    ui.output_progress(text, percent=percent, element_id=element_id)


def output_table(text: str, header: TableHeader, rows: TableRows, element_id: Optional[str] = None) -> None:
    """Display a table.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the table title

    - __header:__ a list of column names

    - __rows:__ the table contents as a list of lists, where each inner list is a row of values

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.output_table:
        text: Cities
        header:
          - Name
          - Country
        rows:
          - - Paris
            - France
          - - London
            - England
          - - Brussels
            - Belgium
    ```
"""
    ui.output_table(text, header=header, data=rows, element_id=element_id)


def output_tabbed_table(text: str, header: TableHeader, tabs: TableTab, element_id: Optional[str] = None) -> None:
    """Display a table with multiple tabs.

!!! question "Minimum Plugin Version: 3.5.0"

??? info "Input"

    - __text:__ the table title

    - __header:__ a list of column names

    - __tabs:__ a dictionary containing tab names & their rows as a list of lists

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Presenting a table with 2 tabs, where each tab contains a table with 3 columns & 1 row.

    ```YAML linenums="1"
    - ppa.ui.output_tabbed_table:
        text: Users
        header:
          - Team
          - Role
          - Location
        tabs:
          John Smith:
            - - Engineering
              - Developer
              - London
          David Bennett:
            - - Engineering
              - Software Tester
              - London
    ```
"""
    ui.output_table(text, header=header, data=tabs, element_id=element_id)


def output_file(
    text: str, data: str, filename: str, type: str = "application/octet-stream", element_id: Optional[str] = None
) -> None:
    """Allow the Task Operator to download a file from the PPA task.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __text:__ text to display with the upload component

    - __data:__ file contents

    - __filename:__ name of the file

    - __type:__ optional MIME type (defaults to `application/octet-stream`)

    - __element_id:__ the ID of the element in the task interface (see [here](../task_interface) for more)

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa.ui.output_file:
        text: Please download the CSV report
        filename: report.csv
      load:
        data: csv_report
    ```

"""
    return ui.output_file(text, data=data, filename=filename, type=type, element_id=element_id)
